import pprint
import typing
import dataclasses
from lxml.etree import _Element  # type: ignore

# utilities for working with optional dependencies
try:
    from attr import has as attr_has
    from attr import fields as attr_fields
    from attr import asdict as attr_asdict
except ImportError:  # pragma: no cover
    attr_has = lambda x: False  # type: ignore # noqa
    attr_fields = lambda x: []  # type: ignore # noqa
    attr_asdict = lambda x: {}  # type: ignore # noqa


def _display_element(obj: _Element) -> str:
    elem_str = f"<{obj.tag} "

    # := if we drop 3.7
    id_str = obj.get("id")
    class_str = obj.get("class")

    if id_str:
        elem_str += f"id='{id_str}'"
    elif class_str:
        elem_str += f"class='{class_str}'"
    else:
        elem_str += " ".join(f"{k}='{v}'" for k, v in obj.attrib.items())

    return f"{elem_str.strip()}> @ line {obj.sourceline}"


def _is_pydantic(obj: typing.Any) -> bool:
    return hasattr(obj, "__fields__") and hasattr(obj, "dict")


def _display(obj: typing.Any) -> str:
    if isinstance(obj, _Element):
        return _display_element(obj)
    else:
        # if there's a dict representation, use that, otherwise str
        try:
            return pprint.pformat(_obj_to_dict(obj))
        except ValueError:
            return str(obj)


def _obj_to_dict(obj: typing.Any) -> typing.Optional[typing.Dict]:
    if obj is None or isinstance(obj, dict):
        return obj
    elif dataclasses.is_dataclass(obj):
        return dataclasses.asdict(obj)
    elif attr_has(obj):
        return attr_asdict(obj)
    elif _is_pydantic(obj):
        return obj.dict()
    else:
        raise ValueError(f"invalid type: {obj} ({type(obj)})")



import re
from typing import Optional, List, Iterator
from lxml.etree import _Element  # type: ignore
from .utils import _display


class SelectorError(ValueError):
    """
    Error raised when a selector's constraint (min_items/max_items, etc.) is not met.
    """

    pass


class Selector:
    """
    Base class implementing **Selector** interface.
    """

    def __init__(
        self,
        *,
        min_items: Optional[int] = 1,
        max_items: Optional[int] = None,
        num_items: Optional[int] = None,
    ):
        self.min_items = min_items
        self.max_items = max_items
        self.num_items = num_items

    def match(
        self,
        element: _Element,
        *,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        num_items: Optional[int] = None,
    ) -> List[_Element]:
        """
        Return all matches of the given selector within `element`.

        If the number of elements matched is outside the prescribed boundaries, a
        `SelectorError` is raised.

        :param element: The element to match within. When used from within a `Page` will
                        usually be `self.root`.
        :param min_items: A minimum number of items to match.
        :param max_items: A maximum number of items to match.
        :param num_items: An exact number of items to match.
        """
        items = list(self.get_items(element))
        num_items = self.num_items if num_items is None else num_items
        max_items = self.max_items if max_items is None else max_items
        min_items = self.min_items if min_items is None else min_items

        if num_items is not None and len(items) != num_items:
            raise SelectorError(
                f"{self} on {_display(element)} got {len(items)} results, "
                f"expected {num_items}"
            )
        if min_items is not None and len(items) < min_items:
            raise SelectorError(
                f"{self} on {_display(element)} got {len(items)} results, "
                f"expected at least {min_items}"
            )
        if max_items is not None and len(items) > max_items:
            raise SelectorError(
                f"{self} on {_display(element)} got {len(items)} results, "
                f"expected at most {max_items}"
            )

        return items

    def match_one(self, element: _Element) -> _Element:
        """
        Return exactly one match.

        :param element: Element to search within.
        """
        return self.match(element, num_items=1)[0]

    def get_items(self, element: _Element) -> Iterator[_Element]:  # pragma: no cover
        raise NotImplementedError()


class XPath(Selector):
    def __init__(
        self,
        xpath: str,
        *,
        min_items: Optional[int] = 1,
        max_items: Optional[int] = None,
        num_items: Optional[int] = None,
    ):
        """
        Utilize [XPath](https://en.wikipedia.org/wiki/XPath#Examples) selectors.

        :param xpath: XPath expression to use.
        :param min_items: A minimum number of items to match.
        :param max_items: A maximum number of items to match.
        :param num_items: An exact number of items to match.
        """
        super().__init__(min_items=min_items, max_items=max_items, num_items=num_items)
        self.xpath = xpath

    def get_items(self, element: _Element) -> Iterator[_Element]:
        yield from element.xpath(self.xpath)

    def __str__(self) -> str:  # pragma: no cover
        return f"XPath({self.xpath})"


class SimilarLink(Selector):
    def __init__(
        self,
        pattern: str,
        *,
        min_items: Optional[int] = 1,
        max_items: Optional[int] = None,
        num_items: Optional[int] = None,
    ):
        """
        Match links that fit a provided pattern.

        :param pattern: Regular expression for link hrefs.
        :param min_items: A minimum number of items to match.
        :param max_items: A maximum number of items to match.
        :param num_items: An exact number of items to match.
        """
        super().__init__(min_items=min_items, max_items=max_items, num_items=num_items)
        self.pattern = re.compile(pattern)

    def get_items(self, element: _Element) -> Iterator[_Element]:
        seen = set()
        for element in element.xpath("//a"):
            href = element.get("href")
            if (
                href
                and href not in seen
                and self.pattern.match(element.get("href", ""))
            ):
                yield element
                seen.add(href)

    def __str__(self) -> str:  # pragma: no cover
        return f"SimilarLink({self.pattern})"


class CSS(Selector):
    def __init__(
        self,
        css_selector: str,
        *,
        min_items: Optional[int] = 1,
        max_items: Optional[int] = None,
        num_items: Optional[int] = None,
    ):
        """
        Utilize CSS-style selectors.

        :param css_selector: CSS selector expression to use.
        :param min_items: A minimum number of items to match.
        :param max_items: A maximum number of items to match.
        :param num_items: An exact number of items to match.
        """
        super().__init__(min_items=min_items, max_items=max_items, num_items=num_items)
        self.css_selector = css_selector

    def get_items(self, element: _Element) -> Iterator[_Element]:
        yield from element.cssselect(self.css_selector)

    def __str__(self) -> str:  # pragma: no cover
        return f"CSS({self.css_selector})"



import dataclasses
import datetime
import functools
import importlib
import inspect
import json
import logging
import sys
import typing
import uuid
from pathlib import Path
from types import ModuleType
import lxml.html  # type: ignore
import click
from scrapelib import Scraper, SQLiteCache
from .utils import _display, _obj_to_dict, attr_has, attr_fields
from .sources import URL, Source
from .pages import Page, ListPage


VERSION = "0.8.7"


def scraper_params(func: typing.Callable) -> typing.Callable:
    @functools.wraps(func)
    @click.option(
        "-ua",
        "--user-agent",
        default=f"spatula {VERSION}",
        help="override default user-agent",
    )
    @click.option("--rpm", default=60, help="set requests per minute (default: 60)")
    @click.option(
        "--timeout", default=5, help="set HTTP request timeout in seconds (default: 5)"
    )
    @click.option(
        "--verify/--no-verify", default=True, help="control verification of SSL certs"
    )
    @click.option(
        "--retries",
        default=0,
        help="configure how many retries to perform on HTTP request error (default: 0)",
    )
    @click.option(
        "--retry-wait",
        default=10,
        help="configure how many seconds to wait on HTTP request error (default: 10)",
    )
    @click.option(
        "-H",
        "--header",
        help="add a header to all requests. example format: 'Accept: application/json'",
        multiple=True,
    )
    @click.option(
        "-v",
        "--verbosity",
        help="override default verbosity for command (0-3)",
        default=-1,
    )
    @click.option(
        "--fastmode",
        help="use a cache to avoid making unnecessary requests",
        is_flag=True,
    )
    def newfunc(
        header: typing.List[str],
        retries: int,
        retry_wait: int,
        rpm: int,
        timeout: int,
        user_agent: str,
        verbosity: int,
        verify: bool,
        fastmode: bool,
        **kwargs: str,
    ) -> None:
        scraper = Scraper(
            requests_per_minute=rpm,
            retry_attempts=retries,
            retry_wait_seconds=retry_wait,
            verify=verify,
        )
        scraper.timeout = timeout
        scraper.user_agent = user_agent
        # only update headers, don't overwrite defaults
        scraper.headers.update(
            {k.strip(): v.strip() for k, v in [h.split(":") for h in header]}
        )
        if fastmode:
            scraper.cache_storage = SQLiteCache("spatula-cache.db")
            scraper.cache_write_only = False

        if verbosity == -1:
            level = logging.INFO if func.__name__ != "test" else logging.DEBUG
        elif verbosity == 0:  # pragma: no cover
            level = logging.ERROR
        elif verbosity == 1:  # pragma: no cover
            level = logging.INFO
        elif verbosity >= 2:  # pragma: no cover
            level = logging.DEBUG

        if verbosity < 3:
            # replace parent library logging
            logging.getLogger("scrapelib").setLevel(logging.ERROR)
            logging.getLogger("urllib3").setLevel(logging.WARNING)
        logging.basicConfig(level=level)

        return func(**kwargs, scraper=scraper)

    return newfunc


def import_mod(name: str) -> typing.Union[ModuleType, type]:
    try:
        return importlib.import_module(name)
    except ImportError:
        # no need to try again
        if "." in sys.path:
            raise
        logging.getLogger("spatula").debug("appending current directory to PYTHONPATH")
        sys.path.append(".")
        return importlib.import_module(name)


def get_page_class(dotted_name: str) -> type:
    mod_name, cls_name = dotted_name.rsplit(".", 1)
    mod = import_mod(mod_name)
    Cls = getattr(mod, cls_name)
    return Cls


def get_dump_function(
    dotted_name: str,
) -> typing.Callable[[typing.Optional[dict], typing.IO], None]:
    mod_name, func_name = dotted_name.rsplit(".", 1)
    mod = import_mod(mod_name)
    func = getattr(mod, func_name)
    return func


def get_pages_from_module(dotted_name: str) -> typing.List[type]:
    mod = import_mod(dotted_name)
    pages = set()
    base_pages = set()
    for name, member in inspect.getmembers(mod):
        try:
            if issubclass(member, ListPage):
                pages.add(member)
                base_pages.update(member.__mro__[1:])
        except TypeError:
            pass
    return list(pages - base_pages)


def get_pages(dotted_name: str, source: typing.Optional[str]) -> typing.List[Page]:
    try:
        pages = [Cls() for Cls in get_pages_from_module(dotted_name)]
        if pages:
            logging.getLogger("spatula").debug(f"found pages: {pages}")
            return pages
        else:
            logging.getLogger("spatula").error(
                f"found no list pages in module {dotted_name}"
            )
            sys.exit(1)
    except ImportError:
        # single page
        Cls = get_page_class(dotted_name)
        if isinstance(Cls, Page):
            if source:
                Cls.source = URL(source)
            return [Cls]
        else:
            return [Cls(source=source)]


def get_new_filename(obj: typing.Any) -> str:
    if hasattr(obj, "get_filename"):
        return obj.get_filename()
    else:
        return str(uuid.uuid4())


@click.group()
@click.version_option(version=VERSION)
def cli() -> None:
    pass


@cli.command()
@click.argument("url")
@click.option("-X", "--verb", default="GET", help="set HTTP verb such as POST")
@scraper_params
def shell(url: str, verb: str, scraper: Scraper) -> None:
    """
    Start a session to interact with a particular page.
    """
    try:
        from IPython import embed  # type: ignore
    except ImportError:  # pragma: no cover
        click.secho("shell command requires IPython", fg="red")
        return

    # import selectors so they can be used without import
    from .selectors import SelectorError, XPath, SimilarLink, CSS  # noqa

    resp = scraper.request(verb, url)
    root = lxml.html.fromstring(resp.content)  # noqa
    click.secho(f"spatula {VERSION} shell", fg="blue")
    click.secho("available selectors: CSS, SimilarLink, XPath", fg="blue")
    click.secho("local variables", fg="green")
    click.secho("---------------", fg="green")
    click.secho("url: %s" % url, fg="green")
    click.secho("resp: requests Response instance", fg="green")
    click.secho(f"root: `lxml HTML element` <{root.tag}>", fg="green")
    embed()


def _get_fake_input(Cls: type, data: typing.List[str], interactive: bool) -> typing.Any:
    # build fake input from command line data if present
    fake_input = {}
    for item in data:
        k, v = item.split("=", 1)
        fake_input[k] = v

    if hasattr(Cls, "example_input"):
        example = getattr(Cls, "example_input")
        for k, v in fake_input.items():
            if isinstance(example, dict):
                example[k] = v
            else:
                setattr(example, k, v)
        return example

    input_type = getattr(Cls, "input_type", None)
    if input_type:
        click.secho(f"{Cls.__name__} expects input ({input_type.__name__}): ")
        if dataclasses.is_dataclass(input_type):
            fields = dataclasses.fields(input_type)
        elif attr_has(input_type):  # pragma: no cover
            # ignore type rules here since dataclasses/attr do not share a base
            # but fields will have a name no matter what
            fields = attr_fields(input_type)  # type: ignore
        for field in fields:
            if field.name in fake_input:
                click.secho(f"  {field.name}: {fake_input[field.name]}")
            elif interactive:
                fake_input[field.name] = click.prompt("  " + field.name)
            else:
                dummy_val = f"~{field.name}"
                fake_input[field.name] = dummy_val
                click.secho(f"  {field.name}: {dummy_val}")
        return input_type(**fake_input)
    else:
        return fake_input


@cli.command()
@click.argument("class_name")
@click.option(
    "--interactive/--no-interactive",
    default=False,
    help="Interactively prompt for missing data. (Default: false)",
)
@click.option(
    "-d", "--data", multiple=True, help="Provide input data in name=value pairs."
)
@click.option("-s", "--source", help="Provide (or override) source URL")
@click.option(
    "--pagination/--no-pagination",
    default=True,
    help="Determine whether or not pagination should be followed or one page is "
    "enough for testing. (Default: true)",
)
@click.option(
    "--subpages/--no-subpages",
    default=False,
    help="Determine whether subpages should be scraped. (Default: false)",
)
@scraper_params
def test(
    class_name: str,
    interactive: bool,
    data: typing.List[str],
    source: typing.Optional[str],
    pagination: bool,
    subpages: bool,
    scraper: Scraper,
) -> None:
    """
    Scrape a single page and see output immediately.

    This eases the common cycle of making modifications to a scraper, running a scrape
    (possibly with long-running but irrelevant portions commented out), and comparing
    output to what is expected.

    ``test`` can also be useful for debugging existing scrapers, you can see exactly
    what a single step of the scrape is providing, to help narrow down where
    erroneous data is coming from.

    Example:

    ``` console
    $ spatula test path.to.ClassName --source https://example.com
    ```

    This will run the scraper defined at `path.to.ClassName` against the provided URL.
    """
    Cls = get_page_class(class_name)
    source_obj: typing.Optional[Source] = None

    fake_input = _get_fake_input(Cls, data, interactive)

    # special case for passing a single URL source as a string
    if source:
        source_obj = URL(source)
    if not source_obj and hasattr(Cls, "example_source"):
        source_obj = Cls.example_source  # type: ignore

    if subpages:
        initial_page = Cls(fake_input, source=source_obj)
        for n, item in enumerate(initial_page._to_items(scraper)):
            click.echo(click.style(f"{n+1}: ", fg="green") + _display(item))
    else:
        # a custom loop instead of _to_items so we can avoid subpages
        # we need to do the request-response-next-page loop at least once
        once = True
        num_items = 0
        while source_obj or once:
            once = False
            page = Cls(fake_input, source=source_obj)

            # fetch data after input is handled, since we might need to build the source
            page._fetch_data(scraper)

            result = page.process_page()

            if isinstance(result, typing.Generator):
                for item in result:
                    # use this count instead of enumerate to handle pagination
                    num_items += 1
                    if isinstance(item, Page):
                        click.echo(
                            click.style(f"{num_items}: would continue with ", fg="blue")
                            + _display(item)
                        )
                    else:
                        click.echo(
                            click.style(f"{num_items}: ", fg="green") + _display(item)
                        )
            else:
                click.secho(_display(result))

            # will be None in most cases, existing the loop, otherwise we restart
            source_obj = page.get_next_source()
            if source_obj:
                if pagination:
                    click.secho(
                        f"paginating for {page.__class__.__name__} source={source_obj}",
                        fg="blue",
                    )
                else:
                    click.secho(
                        "pagination disabled: would paginate for "
                        f"{page.__class__.__name__} source={source_obj}",
                        fg="yellow",
                    )
                    break


@cli.command()
@click.argument("initial_page_name")
@click.option(
    "-o", "--output-dir", default=None, help="override default output directory."
)
@click.option("-s", "--source", help="Provide (or override) source URL")
@click.option("--dump", help="Specify dump function", default="json.dump")
@scraper_params
def scrape(
    initial_page_name: str,
    output_dir: str,
    source: typing.Optional[str],
    scraper: Scraper,
    dump: str,
) -> None:
    """
    Run full scrape, and output data to disk.
    """
    # ensure output directory is ready
    if not output_dir:
        dirn = 1
        today = datetime.date.today().strftime("%Y-%m-%d")
        while True:
            try:
                output_path = Path(f"_scrapes/{today}/{dirn:03d}")
                output_path.mkdir(parents=True)
                break
            except FileExistsError:
                dirn += 1
    else:
        output_path = Path(output_dir)
        try:
            output_path.mkdir(parents=True)
        except FileExistsError:
            if len(list(output_path.iterdir())):
                click.secho(f"{output_dir} exists and is not empty", fg="red")
                sys.exit(1)

    dump_func = get_dump_function(dump)
    # actually do the scrape
    count = 0
    pages = get_pages(initial_page_name, source)
    for initial_page in pages:
        for item in initial_page._to_items(scraper):
            filename = output_path / (get_new_filename(item) + ".json")
            data = _obj_to_dict(item)
            with open(filename, "w") as f:
                dump_func(data, f)
            count += 1
    click.secho(f"success: wrote {count} objects to {output_path}", fg="green")


@cli.command()
@click.argument("initial_page_name")
@click.option("-s", "--source", help="Provide (or override) source URL")
@click.option(
    "-o",
    "--output-file",
    default="scout.json",
    help="override default output file [default: scout.json].",
)
@scraper_params
def scout(
    initial_page_name: str,
    output_file: str,
    source: typing.Optional[str],
    scraper: Scraper,
) -> None:
    """
    Run first step of scrape & output data to a JSON file.

    This command is intended to be used to detect at a first approximation whether
    or not a full scrape might need to be run. If the first layer detects any changes
    it is safe to say that the full run will as well.

    This will work in the common case where a new subpage is added or removed.
    Of course in more advanced cases this depends upon the first page being scraped
    (typically a ListPage derivative) surfacing enough information (perhaps a
    last_updated date) to know whether any of the other pages have been scraped.
    """
    initial_pages = get_pages(initial_page_name, source)
    items: typing.List[typing.Any] = []
    for initial_page in initial_pages:
        items.extend(initial_page._to_items(scraper, scout=True))
    with open(output_file, "w") as f:
        json.dump(items, f, indent=2)
    click.secho(f"success: wrote {len(items)} records to {output_file}", fg="green")


if __name__ == "__main__":  # pragma: no cover
    cli()



from .pages import (  # noqa
    Page,
    HtmlPage,
    XmlPage,
    JsonPage,
    PdfPage,
    ListPage,
    CsvListPage,
    ExcelListPage,
    HtmlListPage,
    JsonListPage,
    XmlListPage,
    MissingSourceError,
    HandledError,
    SkipItem,
)
from .selectors import SelectorError, Selector, XPath, SimilarLink, CSS  # noqa
from .sources import Source, URL, NullSource  # noqa



import io
import csv
import tempfile
import subprocess
import logging
import warnings
import typing
import scrapelib
import lxml.html  # type: ignore
from openpyxl import load_workbook  # type: ignore
from .sources import Source, URL
from .utils import _obj_to_dict


def _to_scout_result(result: typing.Any) -> typing.Dict[str, typing.Any]:
    _next: typing.Optional[str]
    if isinstance(result, Page):
        data = _obj_to_dict(result.input)
        _next = f"{result.__class__.__name__} source={result.source}"
    else:
        data = _obj_to_dict(result)
        _next = None
    return {
        "data": data,
        "__next__": _next,
    }


class SkipItem(Exception):
    """
    To be raised to skip processing of the current item & continue with the next item.

    Example:
    ``` python
    class SomeListPage(HtmlListPage):
        def process_item(self, item):
            if item.name == "Vacant":
                raise SkipItem("vacant")
            # do normal processing here
    ```

    Or, if the skip logic needs to live within a detail Page:
    ``` python
    class SomeDetailPage(HtmlPage):
        def process_page(self):
            if self.input.name == "Vacant":
                raise SkipItem("vacant")
            # do normal processing here
    ```
    """

    def __init__(self, msg: str):
        super().__init__(msg)


class MissingSourceError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class HandledError(Exception):
    def __init__(self, exc: Exception):
        super().__init__(exc)


class Page:
    """
    Base class for all *Page* scrapers, used for scraping information from a single type of page.

    For details on how these methods are called, it may be helpful to read
    [Anatomy of a Scrape](anatomy-of-a-scrape.md).

    **Attributes**

    `source`
    :   Can be set on subclasses of `Page` to define the initial HTTP request
        that the page will handle in its `process_response` method.

        For simple GET requests, `source` can be a string.
        `URL` should be used for more advanced use cases.

    `response`
    :   [`requests.Response`](https://docs.python-requests.org/en/master/api/#requests.Response)
        object available if access is needed to the raw response for any reason.

    `input`
    :   Instance of data being passed upon instantiation of this page.
        Must be of type `input_type`.

    `input_type`
    :   `dataclass`, `attrs` class, or `pydantic` model.
        If set will be used to prompt for and/or validate `self.input`

    `example_input`
    :   Instance of `input_type` to be used when invoking `spatula test`.

    `example_source`
    :   Source to fetch when invoking `spatula test`.

    `dependencies`
    :   Dictionary mapping of names to `Page` objects that will be available before `process_page`.

        For example:
        ``` python

        class EmployeeDetail(HtmlPage):
            dependencies = {"awards": AwardsPage()}
        ```

        Means that before `EmployeeDetail.process_page` is called, it is guaranteed to have the
        output from `AwardsPage` available as `self.awards`.

        See [Specifying Dependencies](advanced-techniques.md#specifying-dependencies) for
        a more detailed explanation.

    **Methods**
    """

    source: typing.Union[None, str, Source] = None
    dependencies: typing.Dict[str, "Page"] = {}
    _cached_dependencies: typing.Dict[str, typing.Any] = {}

    def _fetch_data(self, scraper: scrapelib.Scraper) -> None:
        """
        ensure that the page has all of its data, this is guaranteed to be called
        exactly once before process_page is invoked
        """
        # process dependencies first
        for key, dep in self.dependencies.items():
            use_cache = False
            if isinstance(dep, type):
                dep = dep(self.input)
            else:
                use_cache = True

            if key in self._cached_dependencies:
                setattr(self, key, self._cached_dependencies[key])
            else:
                dep._fetch_data(scraper)
                page_result = dep.process_page()
                setattr(self, key, page_result)
                if use_cache:
                    self._cached_dependencies[key] = page_result

        if not self.source:
            try:
                self.source = self.get_source_from_input()
            except NotImplementedError:
                raise MissingSourceError(
                    f"{self.__class__.__name__} has no source or get_source_from_input"
                )
        if isinstance(self.source, str):
            self.source = URL(self.source)
        # at this point self.source is indeed a Source
        self.logger.info(f"fetching {self.source}")
        try:
            self.response = self.source.get_response(scraper)  # type: ignore
            if getattr(self.response, "fromcache", None):
                self.logger.debug(f"retrieved {self.source} from cache")
        except scrapelib.HTTPError as e:
            self.process_error_response(e)
            raise HandledError(e)
        else:
            self.postprocess_response()

    def _paginate(
        self, scraper: scrapelib.Scraper, scout: bool
    ) -> typing.Iterable[typing.Any]:
        next_source = self.get_next_source()
        if next_source:
            # instantiate the same class with same input, but increment the source
            next_page = type(self)(self.input, source=next_source)
            yield from next_page._to_items(scraper, scout=scout)

    def _to_items(
        self, scraper: scrapelib.Scraper, *, scout: bool = False
    ) -> typing.Iterable[typing.Any]:
        # fetch data for a page, and then call the process_page entrypoint
        try:
            self._fetch_data(scraper)
        except HandledError:
            # ok to proceed, but nothing left to do with this page
            yield from self._paginate(scraper, scout)
            return
        try:
            result = self.process_page()
        except SkipItem as e:
            # a detail page can raise SkipItem, which means no further processing of
            # that detail page (as there is no result)
            self.logger.info(f"SkipItem: {e}")
            return

        # if we got back a generator, we need to process each result
        if isinstance(result, typing.Generator):
            # each item yielded might be a Page or an end-result
            for item in result:
                if scout:
                    yield _to_scout_result(item)
                elif isinstance(item, Page):
                    yield from item._to_items(scraper)
                else:
                    yield item
        elif scout:
            yield _to_scout_result(result)
        elif isinstance(result, Page):
            # single Page result, recurse deeper
            yield from result._to_items(scraper)
        else:
            # end-result, just return as-is
            yield result

        # check for next page
        yield from self._paginate(scraper, scout)

    def __init__(
        self,
        input_val: typing.Any = None,
        *,
        source: typing.Union[None, str, Source] = None,
    ):
        self.input = input_val
        # allow possibility to override default source, useful during dev
        if source:
            self.source = source
        self.logger = logging.getLogger(
            self.__class__.__module__ + "." + self.__class__.__name__
        )

    def __str__(self) -> str:
        s = f"{self.__class__.__name__}("
        if self.input:
            s += f"input={self.input} "
        if self.source:
            s += f"source={self.source}"
        s += ")"
        return s

    def do_scrape(
        self, scraper: typing.Optional[scrapelib.Scraper] = None
    ) -> typing.Iterable[typing.Any]:
        """
        yield results from this page and any subpages

        :param scraper: Optional `scrapelib.Scraper` instance to use for running scrape.
        :returns: Generator yielding results from the scrape.
        """
        if scraper is None:
            scraper = scrapelib.Scraper()
        yield from self._to_items(scraper)

    def get_source_from_input(self) -> typing.Union[None, str, Source]:
        """
        To be overridden.

        Convert `self.input` to a `Source` object such as a `URL`.
        """
        raise NotImplementedError()

    def postprocess_response(self) -> None:
        """
        To be overridden.

        This is called after source.get_response but before self.process_page.
        """
        pass

    def process_error_response(self, exception: Exception) -> None:
        """
        To be overridden.

        This is called after source.get_response if an exception is raised.
        """
        raise exception

    def process_page(self) -> typing.Any:
        """
        To be overridden.

        Return data extracted from this page and this page alone.
        """
        raise NotImplementedError()

    def get_next_source(self) -> typing.Union[None, str, Source]:
        """
        To be overriden for paginated pages.

        Return a URL or valid source to fetch the next page, None if there isn't one.
        """
        return None


class HtmlPage(Page):
    """
    Page that automatically handles parsing and normalizing links in an HTML response.

    **Attributes**

    `root`
    :   [`lxml.etree.Element`](https://lxml.de/api/lxml.etree._Element-class.html)
    object representing the root element (e.g. `<html>`) on the page.

        Can use the normal lxml methods (such as `cssselect` and `getchildren`), or
        use this element as the target of a `Selector` subclass.
    """

    def postprocess_response(self) -> None:
        self.root = lxml.html.fromstring(self.response.content)
        if hasattr(self.source, "url"):
            self.root.make_links_absolute(self.source.url)  # type: ignore


class XmlPage(Page):
    """
    Page that automatically handles parsing a XML response.

    **Attributes**

    `root`
    :   [`lxml.etree.Element`](https://lxml.de/api/lxml.etree._Element-class.html)
    object representing the root XML element on the page.
    """

    def postprocess_response(self) -> None:
        self.root = lxml.etree.fromstring(self.response.content)


class JsonPage(Page):
    """
    Page that automatically handles parsing a JSON response.

    **Attributes**

    `data`
    :   JSON data from response.  (same as `self.response.json()`)
    """

    def postprocess_response(self) -> None:
        self.data = self.response.json()


class PdfPage(Page):  # pragma: no cover
    """
    Page that automatically handles converting a PDF response to text using `pdftotext`.

    **Attributes**

    `preserve_layout`
    :   set to `True` on derived class if you want the conversion function to use pdftotext's
        -layout option to attempt to preserve the layout of text.
        (`False` by default)

    `text`
    :   UTF8 text extracted by pdftotext.
    """

    preserve_layout = False

    def postprocess_response(self) -> None:
        with tempfile.NamedTemporaryFile() as temp:
            temp.write(self.response.content)
            temp.flush()
            temp.seek(0)

            if self.preserve_layout:
                command = ["pdftotext", "-layout", temp.name, "-"]
            else:
                command = ["pdftotext", temp.name, "-"]

            try:
                pipe = subprocess.Popen(
                    command, stdout=subprocess.PIPE, close_fds=True
                ).stdout
            except OSError as e:
                raise EnvironmentError(
                    f"error running pdftotext, missing executable? [{e}]"
                )
            if pipe is None:
                raise EnvironmentError("no stdout from pdftotext")
            else:
                data = pipe.read()
                pipe.close()
        self.text = data.decode("utf8")


class ListPage(Page):
    """
    Base class for common pattern of extracting many homogenous items from one page.

    Instead of overriding `process_response`, subclasses should provide a `process_item`.

    **Methods**
    """

    def skip(self, msg: str = "") -> None:  # pragma: no cover
        warnings.warn(
            "self.skip is deprecated and will be removed in 0.9, raise SkipItem instead",
            DeprecationWarning,
        )
        raise SkipItem(msg)

    def _process_or_skip_loop(
        self, iterable: typing.Iterable
    ) -> typing.Iterable[typing.Any]:
        for item in iterable:
            try:
                item = self.process_item(item)
            except SkipItem as e:
                self.logger.info(f"SkipItem: {e}")
                continue
            yield item

    def process_item(self, item: typing.Any) -> typing.Any:
        """
        To be overridden.

        Called once per subitem on page, as defined by the particular subclass being used.

        Should return data extracted from the `item`.

        If [`SkipItem`](#SkipItem) is raised, `process_item` will continue to be called with
        the next item in the stream.
        """
        warnings.warn(f"process_item not overridden on {self.__class__.__name__}")
        return item


class CsvListPage(ListPage):
    """
    Processes each row in a CSV (after the first, assumed to be headers) as an item
    with `process_item`.
    """

    def postprocess_response(self) -> None:
        self.reader = csv.DictReader(io.StringIO(self.response.text))

    def process_page(self) -> typing.Iterable[typing.Any]:
        yield from self._process_or_skip_loop(self.reader)


class ExcelListPage(ListPage):  # pragma: no cover
    """
    Processes each row in an Excel file as an item with `process_item`.
    """

    def postprocess_response(self) -> None:
        workbook = load_workbook(io.BytesIO(self.response.content))
        # TODO: allow selecting this with a class property
        self.worksheet = workbook.active

    def process_page(self) -> typing.Iterable[typing.Any]:
        yield from self._process_or_skip_loop(self.worksheet.values)


class LxmlListPage(ListPage):
    """
    Base class for XML and HTML subclasses below, only difference is which
    parser is used.

    Simplification for pages that get a list of items and process them.

    When overriding the class, instead of providing process_page, one must only provide
    a selector and a process_item function.
    """

    selector = None

    def process_page(self) -> typing.Iterable[typing.Any]:
        if not self.selector:
            raise NotImplementedError("must either provide selector or override scrape")
        items = self.selector.match(self.root)
        yield from self._process_or_skip_loop(items)


class HtmlListPage(LxmlListPage, HtmlPage):
    """
    Selects homogenous items from HTML page using `selector` and passes them to `process_item`.

    **Attributes**

    `selector`
    :   `Selector` subclass which matches list of homogenous elements to process.  (e.g. `CSS("tbody tr")`)
    """

    pass


class XmlListPage(LxmlListPage, XmlPage):
    """
    Selects homogenous items from XML document using `selector` and passes them to `process_item`.

    **Attributes**

    `selector`
    :   `Selector` subclass which matches list of homogenous elements to process.  (e.g. `XPath("//item")`)
    """

    pass


class JsonListPage(ListPage, JsonPage):
    """
    Processes each element in a JSON list as an item with `process_item`.
    """

    def process_page(self) -> typing.Iterable[typing.Any]:
        yield from self._process_or_skip_loop(self.data)



from typing import Optional
import requests
import scrapelib


class Source:
    pass


class URL(Source):
    def __init__(
        self,
        url: str,
        method: str = "GET",
        data: dict = None,
        headers: dict = None,
        verify: bool = True,
        timeout: Optional[float] = None,
    ):
        """
        Defines a resource to fetch via URL, particularly useful for handling non-GET
        requests.

        :param url: URL to fetch
        :param method: HTTP method to use, defaults to "GET"
        :param data: POST data to include in request body.
        :param headers: dictionary of HTTP headers to set for the request.
        :param verify: bool indicating whether or not to verify SSL certificates for request, defaults to True
        :param timeout: HTTP(S) timeout in seconds
        """

        self.url = url
        self.method = method
        self.data = data
        self.headers = headers
        self.verify = verify
        self.timeout = timeout

    def get_response(
        self, scraper: scrapelib.Scraper
    ) -> Optional[requests.models.Response]:
        return scraper.request(
            method=self.method,
            url=self.url,
            data=self.data,
            headers=self.headers,
            verify=self.verify,
            timeout=self.timeout,
        )

    def __str__(self) -> str:
        return self.url


class NullSource(Source):
    """
    Special class to set as a page's `source` to indicate no HTTP request needs
    to be performed.
    """

    def get_response(
        self, scraper: scrapelib.Scraper
    ) -> Optional[requests.models.Response]:
        return None

    def __str__(self) -> str:
        return self.__class__.__name__



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['spatula']

package_data = \
{'': ['*']}

install_requires = \
['attrs[attrs]>=20.3.0,<21.0.0',
 'click>=7.1.2,<8.0.0',
 'cssselect>=1.1.0,<2.0.0',
 'ipython[shell]>=7.19.0,<8.0.0',
 'lxml>=4.6.2,<5.0.0',
 'openpyxl>=3.0.6,<4.0.0',
 'scrapelib>=2.0.6,<3.0.0']

entry_points = \
{'console_scripts': ['spatula = spatula.cli:cli']}

setup_kwargs = {
    'name': 'spatula',
    'version': '0.8.7',
    'description': 'A modern Python library for writing maintainable web scrapers.',
    'long_description': '# Overview\n\n*spatula* is a modern Python library for writing maintainable web scrapers.\n\nSource: [https://github.com/jamesturk/spatula](https://github.com/jamesturk/spatula)\n\nDocumentation: [https://jamesturk.github.io/spatula/](https://jamesturk.github.io/spatula/)\n\nIssues: [https://github.com/jamesturk/spatula/issues](https://github.com/jamesturk/spatula/issues)\n\n[![PyPI badge](https://badge.fury.io/py/spatula.svg)](https://badge.fury.io/py/spatula)\n[![Test badge](https://github.com/jamesturk/spatula/workflows/Test%20&%20Lint/badge.svg)](https://github.com/jamesturk/spatula/actions?query=workflow%3A%22Test+%26+Lint%22)\n\n## Features\n\n- **Page-oriented design**: Encourages writing understandable & maintainable scrapers.\n- **Not Just HTML**: Provides built in [handlers for common data formats](https://jamesturk.github.io/spatula/reference/#pages) including CSV, JSON, XML, PDF, and Excel.  Or write your own.\n- **Fast HTML parsing**: Uses `lxml.html` for fast, consistent, and reliable parsing of HTML.\n- **Flexible Data Model Support**: Compatible with `dataclasses`, `attrs`, `pydantic`, or bring your own data model classes for storing & validating your scraped data.\n- **CLI Tools**: Offers several [CLI utilities](https://jamesturk.github.io/spatula/cli/) that can help streamline development & testing cycle.\n- **Fully Typed**: Makes full use of Python 3 type annotations.\n',
    'author': 'James Turk',
    'author_email': 'dev@jamesturk.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jamesturk/spatula/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
